/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/

#ifndef cl_array_h
#define cl_array_h

#include <assert.h>

template <class T>
class cl_Array
{
	protected:
		T *data;
		unsigned long size;
		unsigned long alloc_size;
	public:
		cl_Array(void) : data((T*)0),size(0),alloc_size(0){};
		
		cl_Array(unsigned long items) : data((T*)0),size(items),alloc_size(items)
		{
			if(items) data = new T[items];
		};
		
		cl_Array(const cl_Array &rv);
		
		~cl_Array(void)
		{
			if(data) delete []data;
		}
		
		// create a _new_ array with the specified size, 
		// discard any previous data.
		int init(unsigned long items);
		
		void clear(void) 
		{
			init(0);
		}
	
		// set the size of the array to the specified value and
		// discard any previous data.
		int resize(unsigned long newlen) 
		{
			if (newlen <= alloc_size) 
			{
				size = newlen;
				return 0;
			}
			return init(newlen);
		}
	
		// add "newitems" free slots to the end of the existing array
		int extend(unsigned long newitems);
	
		// cut the specified range from the existing array
		void cut(unsigned long begin, unsigned long end);
		
		void cut_tail(unsigned long num) 
		{
			cut((num > size)? 0 : size-num, size);
		}
		
		void cut_head(unsigned long num) 
		{
			cut(0, (num > size)? size : num);
		}
	
		// initialize the size and contents of the array from the 
		// specified memory location and length. 
		int set_to(const T* idata, unsigned long ilen);
	
		// append objects from the specified memory location
		// to the existing array
		int append(const T* newdata, unsigned long newlen);
		
		T& at(int iv) const 
		{
			assert((unsigned long)iv < size && iv >= 0);
			return data[iv];
		};
		
		// operators
		cl_Array& operator= (const cl_Array &rv) 
		{
			set_to(rv.data, rv.size);
			return *this;
		}
	
		cl_Array& operator+= (const cl_Array &rv) 
		{
			append(rv.data, rv.size);
			return *this;
		}
	
		T& operator[] (unsigned long iv) const 
		{
			assert(iv < size);
			return data[iv];
		};

		T& operator[] (unsigned int iv) const 
		{
			assert(iv < size);
			return data[iv];
		};
	
		T& operator[] (int iv) const 
		{
			assert((unsigned long)iv < size && iv >= 0);
			return data[iv];
		};
	
		operator const T* () const 
		{
			return data;
		}

		operator const T* ()
		{
			return data;
		}
	
		operator T* () 
		{
			return data;
		}
	
		// const member functions
		T* get_head(void) const {return data;};
		
		T* get_tail(void) const 
		{
			T* res = data;
			if (!res) return res;
			return res + (size - 1);
		}
		
		const T* get_next(const T* aktp) const 
		{
			assert(aktp >= data && aktp < (data + size));
			return aktp + 1;
		};
		
		const T* get_prev(const T* aktp) const 
		{
			assert(aktp > data && aktp <= (data + size));
			return aktp - 1;
		};
	
		T* get_next(T* aktp) const 
		{
			assert(aktp >= data && aktp < (data + size));
			return aktp + 1;
		};
			
		T* get_prev(T* aktp) const 
		{
			assert(aktp > data && aktp <= (data + size));
			return aktp - 1;
		};
	
		unsigned long length(void) const {return size;}
};	

//##############################################################################

template <class T>
cl_Array<T>::cl_Array(const cl_Array<T> &rv) : data((T*)0), size(rv.size), alloc_size(rv.size) 
{
	if (size) 
	{
		data = new T[size];
		if (data) 
		{
			unsigned long x = size - 1;
			const T* src = rv.data;
			T* dst = data;
			do {
				*dst++ = *src++;
			} while (x--);
		}
	}
}

template <class T>
int cl_Array<T>::init(unsigned long items) 
{
	if (data) delete[] data;
	data = (T*)0;
	size = items;
	alloc_size = items;
	if (items) 
	{
		data = new T[items];
		if (!data) return -1;
	}
	return 0;
}

template <class T>
int cl_Array<T>::set_to(const T* idata, unsigned long ilen) 
{
	
	if (resize(ilen)) return -1;
	
	if (ilen) 
	{
		unsigned long x = ilen - 1;
		const T* src = idata;
		T* dst = data;
		do {
			*dst++ = *src++;
		} while (x--);
	}
	
	return 0;
}

template <class T>
int cl_Array<T>::extend(unsigned long newitems) 
{
	if (newitems == 0) return 0;
	if (alloc_size == 0) return init(newitems);
	
	if (size+newitems <= alloc_size) 
	{
		// no need to alloc new memory, just adjust size
		size += newitems;
		return 0;
	}
	
	// need to alloc new memory
	unsigned long newsize = size+newitems;
	unsigned long new_alloc_size = newsize;
	if (sizeof(T)*new_alloc_size*2 <= 8192) 
	{
		new_alloc_size *= 2; // pre-alloc more space by double-sizing the array
	} 
	else 
	{
		new_alloc_size += 8192/sizeof(T);
	}
	
	T* olddata = data;
	data = new T[new_alloc_size];
	if (!data) 
	{
		data = olddata;
		return -1;
	}
	{
		// copy old contents
		unsigned long x = size;
		if (x) 
		{
			x--;
			const T* src = olddata;
			T* dst = data;
			do {
				*dst++ = *src++;
			} while (x--);
		}
	}
	
	delete[] olddata;
	
	size = newsize;
	alloc_size = new_alloc_size;
	
	return 0;
}

template <class T>
void cl_Array<T>::cut(unsigned long begin, unsigned long end) 
{	
	if (begin > size) 
	{
		return;
	}
	if (end > size) 
	{
		end = size;
	}
	if (end <= begin) return;
		
	// copy contents
	const T* src = data+end;
	T* dst = data+begin;
	unsigned long i;
	for (i = end; i < size; i++) 
	{
		*dst++ = *src++;
	}
	
	size -= end - begin;	
	
	return;
}

template <class T>
int cl_Array<T>::append (const T* newdata, unsigned long newlen) 
{	
	if (newlen) 
	{	
		unsigned long oldsize = size;
		
		if (extend(newlen)) return -1;
		
		unsigned long x = newlen - 1;
		const T* src = newdata;
		T* dst = data+oldsize;
		do {
			*dst++ = *src++;
		} while (x--);
		
	}
	
	return 0;
}

#endif /*cl_array_h*/
