/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/

#ifndef __common_h
#define __common_h

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/statfs.h>

/* prevent loading structure from asm/statfs.h 
   which conflicts with bits/statfs.h */
#define _I386_STATFS_H
#include <linux/fs.h>

#include "cl_string.h"

#define BUFFERSIZE 4096
#define BLOCK(X,Y) ( (int((int(Y)+1)*blocksize)) > (int(X)) ? (int(int(X)-int(int(Y)*blocksize))) : (int(blocksize)) )
typedef int MY_FUNC(const char *,const struct stat *);

extern int blocksize;
extern int verbose;
extern char* dummybuffer;

int io_error(const char *message, bool do_exit)
{
	char buf[1024];
	sprintf(buf, "Drastic I/O error (%s)", message);
	perror(buf);
	if(do_exit) exit(1);
	return(1);
}

int writefile(const char* pathname,int length=BUFFERSIZE,int access=O_CREAT|O_WRONLY,mode_t perm=S_IRUSR|S_IWUSR){

	int fdesc=open(pathname,access,perm);
	if(fdesc==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		

	int counter=0;
	int length_=0;
	int size=0;

	while((length_=BLOCK(length,counter))>0)
	{
		int ret=write(fdesc,dummybuffer,length_);
		if((ret==-1)||(ret!=length_)){
			fprintf(stderr,"Error writing \"%s\", reason: %s\n",pathname,strerror(errno));
			close(fdesc);
			return -1;
		}		
		counter++;
		size+=ret;
	}

	if(close(fdesc)==-1){
		fprintf(stderr,"Error closing \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		
	return size;
}

int readfile(const char* pathname,int length=-1,int access=O_RDONLY, mode_t perm=S_IRUSR|S_IWUSR){

	int fdesc=open((char const*)pathname,access,perm);
	if(fdesc==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		

	int counter=0;
	int length_=0;
	int size=0;

	if(length<0){
		struct stat statbuf;
		if(fstat(fdesc,&statbuf)==-1){
			fprintf(stderr,"Error stating \"%s\", reason: %s\n",pathname,strerror(errno));
			close(fdesc);
			return -1;
		}		
		length=statbuf.st_size;
	}
	
	while((length_=BLOCK(length,counter))>0)
	{
		int ret=read(fdesc,dummybuffer,length_);
		if(ret==-1){
			fprintf(stderr,"Error reading \"%s\", reason: %s\n",pathname,strerror(errno));
			close(fdesc);
			return -1;
		}		
		else if(ret==0) break;
		size+=ret;
		counter++;
		//cout << counter  << endl; 
	}

	if(close(fdesc)==-1){
		fprintf(stderr,"Error closing \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		
	return size;
}

int unlinkfile(const char* pathname){

	if(unlink((char const*)pathname)==-1){
		fprintf(stderr,"Error unlinking \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		

	return 0;
}

int symlinkfile(const char* pathname, const char* newpathname){

	if(symlink((const char*)pathname,(const char*)newpathname)==-1){
		fprintf(stderr,"Error symlinking \"%s\" to \"%s\", reason: %s\n",pathname,newpathname,strerror(errno));
		return -1;
	}		

	return 0;
}


int statfile(const char* pathname,float *res=NULL){
	
	int access=O_RDONLY;
	mode_t perm=S_IRUSR|S_IWUSR;
	int fdesc=open(pathname,access,perm);
	if(fdesc==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		

	struct stat statbuf;
	if(fstat(fdesc,&statbuf)==-1){
		fprintf(stderr,"Error stating \"%s\", reason: %s\n",pathname,strerror(errno));
		close(fdesc);
		return -1;
	}		

	if(res==NULL) return 0;

	*res=1.0;
	int bq=statbuf.st_blksize/512;
	//printf("File size: %ld Bytes\n", statbuf.st_size);
	//printf("Block usage: %ld\n", statbuf.st_blocks/bq);
	//printf("Block size: %ld Bytes\n", statbuf.st_blksize);

	int fib_blk_old=0;
	int counter=0;
	int i;
	for(i=0;i<statbuf.st_blocks/bq;i++){
		int fib_blk=i;
		int	fib_res=ioctl(fdesc,FIBMAP,&fib_blk);
		if(fib_res==-1)
		{
			fprintf(stderr,"Error FIBMAPing \"%s\", reason: %s\n",pathname,strerror(errno));
			return -1;
		}
		//printf("%d %d\n",i,fib_blk);
		if(i==0){
			fib_blk_old=fib_blk;
			continue;
		}
		else if (fib_blk-fib_blk_old==1) counter++;
		fib_blk_old=fib_blk;
	}
	if(i>1){
		*res=(float)counter/(i-1);
		//if(verbose==1) printf("FIBMAP Result for %d blocks: %f\n",i,(float)counter/(i-1));
	}

	if(close(fdesc)==-1){
		fprintf(stderr,"Error closing \"%s\", reason: %s\n",pathname,strerror(errno));
		return -1;
	}		

	return 0;
}

int makedir(const char *pathname,mode_t perm=S_IRUSR|S_IWUSR|S_IXUSR){

	if(mkdir(pathname,perm)==-1){
		fprintf(stderr,"Error making directory \"%s\", reason: %s\n",(char *)pathname,strerror(errno));
		return -1;
	}		
	return 0;
}

int changedir(const char *pathname){

	if(chdir(pathname)==-1){
		fprintf(stderr,"Error changing directory \"%s\", reason: %s\n",(char *)pathname,strerror(errno));
		return -1;
	}		

	return 0;
}

int removedir(const char *pathname){

	if(rmdir(pathname)==-1){
		fprintf(stderr,"Error removing directory \"%s\", reason: %s\n",(char *)pathname,strerror(errno));
		return -1;
	}		

	return 0;
}

int walk_through_dirs(const char *path, MY_FUNC *function)
{
	struct stat statbuf;
	if(lstat(path,&statbuf)==-1){
		fprintf(stderr,"Error stating \"%s\", reason: %s\n",path,strerror(errno));
		return -1;
	}
	
	if((S_ISLNK(statbuf.st_mode)==1)||(S_ISDIR(statbuf.st_mode)==0)) 
		return (function(path,&statbuf)); /*not a dir*/

	int ret=function(path,&statbuf); /*work on dir first and save result*/

	DIR *dirz;
	dirent *direntz;
	if((dirz=opendir(path))==NULL) {
		fprintf(stderr,"Error opening dir \"%s\", reason: %s\n",path,strerror(errno));
		closedir(dirz);
		return (function(path,&statbuf));
	}

	cl_String npath;
	while((direntz=readdir(dirz))!=NULL){
		/* ignore "." and ".." */
		if((strcmp(direntz->d_name,".")!=0)&&(strcmp(direntz->d_name,"..")!=0)){
			npath=path;
			npath+="/";
			npath+=direntz->d_name;
			if(walk_through_dirs((char *)npath,function)!=0) break;
		}
	}
	
	if(closedir(dirz)<0){
		fprintf(stderr,"Error closing dir \"%s\", reason: %s\n",(char *)path,strerror(errno));
		return -1;
	} 
	
	return ret;

}

long countdirentries(const char *path)
{
	DIR *dirz;
	dirent *direntz;
	if((dirz=opendir(path))==NULL) {
		fprintf(stderr,"Error opening dir \"%s\", reason: %s\n",path,strerror(errno));
		closedir(dirz);
		return -1;
	}

	long ret=-2;
	while((direntz=readdir(dirz))!=NULL) ret++;
	
	if(closedir(dirz)<0){
		fprintf(stderr,"Error closing dir \"%s\", reason: %s\n",(char *)path,strerror(errno));
		return -1;
	}

	return ret;
}

#endif
